#ifndef PRINTBENCH_H
#define PRINTBENCH_H

#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>

#include "cpucycles.c"

#ifdef PRINTBENCH_1
static void cpucycles_fprintf_1(int argc, uint64_t results[OP][LOOPS], char *op_str[])
{
  int op, loop;
  FILE *f;

  // sort 'loop'
  for (op = 0; op < OP; op++)
  { qsort(&(results[op][0]),LOOPS,sizeof(uint64_t),cmp_uint64); }

  // print to file
  f = stdout;
  for (op = 0; op < OP; op++)
  { if(argc == 1) { f = fopen(op_str[op], "w"); }
    for (loop = 0; loop < LOOPS-1; loop++)
    { fprintf(f, "%" PRIu64 ", ", results[op][loop]); }
    fprintf(f, "%" PRIu64 "\n", results[op][LOOPS-1]);
    if(argc == 1) { fclose(f); }
  }
}
#endif

#ifdef PRINTBENCH_2
void alloc_2(uint64_t* results[OP][LOOPS], size_t _x)
{
  size_t o, l;
  for (o=0; o<OP; o++)
  { for (l=0; l<LOOPS; l++)
    { results[o][l] = (uint64_t*) malloc(sizeof(uint64_t) * _x); }
  }
}

void free_2(uint64_t* results[OP][LOOPS])
{
  size_t o, l;
  for (o=0; o<OP; o++)
  { for (l=0; l<LOOPS; l++)
    { free(results[o][l]); }
  }
}

static void cpucycles_fprintf_2(int argc, uint64_t* results[OP][LOOPS], char *op_str[])
{
  int op, loop, r;
  uint64_t len;
  double cpb;
  FILE *f;
  uint64_t temp[LOOPS];

  // sort 'loop'
  for (len = MININBYTES, r=0; len <= MAXINBYTES; len = inc_in(len), r +=1)
  { for (op = 0; op < OP; op++)
    { for (loop = 0; loop < LOOPS; loop++)
      { temp[loop] = results[op][loop][r]; } 
      qsort(temp,LOOPS,sizeof(uint64_t),cmp_uint64);
      for (loop = 0; loop < LOOPS; loop++)
      { results[op][loop][r] = temp[loop]; } 
    }
  }

  // print to file
  f = stdout;
  for (op = 0; op < OP; op++)
  { if(argc == 1) { f = fopen(op_str[op], "w"); }
    loop = 0;
    for (len = MININBYTES, r=0; len <= MAXINBYTES; len = inc_in(len), r +=1)
    { if(len == 0)
      { fprintf(f, "%" PRIu64 ", %" PRIu64 ", undef, ", len, results[op][LOOPS>>1][r]); }
      else
      { cpb = ((double)results[op][loop][r]) / ((double)len);
        fprintf(f, "%" PRIu64 ", %" PRIu64 ", %.2f, ", len, results[op][LOOPS>>1][r], cpb );
      }
      for (loop = 0; loop < LOOPS-1; loop++)
      { fprintf(f, "%" PRIu64 ", ", results[op][loop][r]); }
      fprintf(f, "%" PRIu64 "\n", results[op][LOOPS-1][r]);
    }
    if(argc == 1) { fclose(f); }
  }
}
#endif

#ifdef PRINTBENCH_3
void alloc_3(uint64_t** results[OP][LOOPS], size_t _x, size_t _y)
{
  size_t o, l, x;
  for (o=0; o<OP; o++)
  { for (l=0; l<LOOPS; l++)
    { results[o][l] = (uint64_t**) malloc(sizeof(uint64_t*) * _x);
      for (x=0; x<_x; x++)
      { results[o][l][x] = (uint64_t*) malloc(sizeof(uint64_t) * _y); }
    }
  }
}

void free_3(uint64_t** results[OP][LOOPS], size_t _x)
{
  size_t o, l, x;
  for (o=0; o<OP; o++)
  { for (l=0; l<LOOPS; l++)
    { for (x=0; x<_x; x++)
      { free(results[o][l][x]); }
      free(results[o][l]);
    }
  }
}

static void cpucycles_fprintf_3(int argc, uint64_t** results[OP][LOOPS], char *op_str[])
{
  int op, loop, r0, r1;
  uint64_t outlen, inlen; //, min;
  double cpb;
  FILE *f;
  uint64_t temp[LOOPS];

  // sort 'loop'
  for (outlen = MINOUTBYTES, r0 = 0; outlen <= MAXOUTBYTES; outlen = inc_out(outlen), r0 += 1)
  { for (inlen = MININBYTES, r1 = 0; inlen <= MAXINBYTES; inlen = inc_in(inlen), r1 += 1)
    { for (op = 0; op < OP; op++)
      { for (loop = 0; loop < LOOPS; loop++)
        { temp[loop] = results[op][loop][r0][r1]; } 
        qsort(temp,LOOPS,sizeof(uint64_t),cmp_uint64);
        for (loop = 0; loop < LOOPS; loop++)
        { results[op][loop][r0][r1] = temp[loop]; }
      }
    }
  }

  // print to file
  f = stdout;
  for (op = 0; op < OP; op++)
  { if(argc == 1) { f = fopen(op_str[op], "w"); }
    loop = 0;
    for (outlen = MINOUTBYTES, r0 = 0; outlen <= MAXOUTBYTES; outlen = inc_out(outlen), r0 += 1)
    { for (inlen = MININBYTES, r1 = 0; inlen <= MAXINBYTES; inlen = inc_in(inlen), r1 += 1)
      { if(inlen == 0)
        { fprintf(f, "%" PRIu64 ", %" PRIu64 ", %" PRIu64 ", undef, ", outlen, inlen, results[op][LOOPS>>1][r0][r1]); }
        else
        { cpb = ((double)results[op][LOOPS>>1][r0][r1]) / ((double)inlen); // cycles per input byte
          fprintf(f, "%" PRIu64 ", %" PRIu64 ", %" PRIu64 ", %.2f, ", outlen, inlen, results[op][LOOPS>>1][r0][r1], cpb );
        }
        for (loop = 0; loop < LOOPS-1; loop++)
        { fprintf(f, "%" PRIu64 ", ", results[op][loop][r0][r1]); }
        fprintf(f, "%" PRIu64 "\n", results[op][LOOPS-1][r0][r1]);
      }
    }
    if(argc == 1) { fclose(f); }
  }
}
#endif

#endif

