# Overview

This artifact allows you to replicate the paper results. We tested the artifact under a Linux system. Next, we detail:

1. how to run the ChaCha demo
2. how to inspect the benchmark results that are shipped with this artifact and used in the paper
3. how to run the benchmarks without building/installing the Jasmin compiler
4. how to build the Jasmin compiler, remove the distributed assembly files, build them again, run the benchmarks, and inspect the results 


# ChaCha Demo

The demo is included under the `chacha20-demo/` directory, which contains a `README.md`. You will need `docker-compose`, `python3`, and to install some packages. You can find detailed information there. The expected result is something similar to the following:

```
zeroization/chacha20-demo/exploit$ python3 decrypt.py localhost 55432
[+] Opening connection to localhost on port 55432: Done
ciphertext: bbfa813e7543145403da0fef60d6892277fb847f5576f153f4c74a40f6a2aa62
recovered partial key: b91605dc27aa6a8f85525173b22c149a9986bc5cbc72915e106b9571a40db1b1
recovered full key: b91605dc27aa6a8f85525173b22c149a9986bc5cbc72915e106b9571a40db1b1
recovered nonce: 8227d5cb4a360f02
Flag:  b'CSR{InsaneInFRAUXIPAUXIBBINGOOO\x00'
[*] Closed connection to localhost port 55432
```

After running the example, don't forget to kill/remove the container, for instance:
```
$ docker ps -a
$ docker kill stapelspass_xinetd_1
$ docker rm stapelspass_xinetd_1
$ docker image rm stapelspass_xinetd
```


# Quickly inspecting some existing results

In the directory `zeroization/table/`, you can find a `Makefile` to build `zeroization/table/main.pdf` file. In `zeroization/table/`, run the following command (requires `pdflatex`):
```
$ make
```

Open `main.pdf` if you wish to do so. As additional information (which might be helpful for the next steps), the tables from `tables/main.pdf` are defined in `tables/tables.tex`, which uses latex commands defined in `tables/data/{cycles,lines,stack}.tex`. Files under `table/data/` are generated by scripts under `zeroization/scripts`. For more information, read the rule `compare` in `zeroization/Makefile/`.


# Running the benchmarks without building the Jasmin compiler

In the root directory of this artifact, `zeroization`, to run the benchmarks (if you wish to get stable results, disable Turbo Boost and Hyper-Threading, and check if the machine is somewhat idle):
```
$ make bench 
$ make compare
```

The first command will create a series of binaries named `bench` under `libjade-*/bench/bin` (one for each assembly file; they are not deleted once the benchmarking ends; you can run them independently afterwards). This process usually takes between 10 and 15 minutes, depending on the CPU. Next to the `bench` executables, there are `*.csv` files containing the results. To find the location of all `.csv` files, you can run the following:
```
$ find libjade-*/bench/bin -name "*.csv"
```

The second command, `make compare`, creates the directories `tables/results/latest/0_baseline_vs_loop` and `tables/results/latest/0_baseline_vs_unrolled`, which will contain the files with the comparison results, and then calls a series of scripts to ensure that files under `table/data/` get updated. The last command executed by `make compare` tries to open `tables/main.pdf`.

Running the benchmarks without building the Jasmin compiler is possible in this artifact because we include the assembly files. If you wish to do some manual inspection of those, you can find the location of all `.s` files with the following command:
```
$ find libjade-* -name "*.s"
```


# Building the Jasmin compiler and compiling the code

Perhaps the most convenient way to build the Jasmin compiler is by using `nix-shell`. If you already have `nix-shell` installed in your system, or you don't mind installing it, you can try to run the following command:
```
$ cd zeroization/jasmin
$ nix-shell --command "(cd compiler && make CIL && make)"
$ sudo install -D compiler/jasminc /usr/local/bin/
```

Where you can replace the `/usr/local/bin/` for some directory of your choice. If you prefer to use docker (it takes around 10 minutes to build): 
```
$ cd zeroization/scripts/
$ docker build -t zeroization -f Dockerfile ../
$ docker run --name zeroization_1 -it zeroization bash
```

At this point, you can start by cleaning the artifact assembly files and benchmarking results under `libjade-*` directories (it takes around 2.5 minutes to build the assembly files):
```
$ ls 
Makefile  README.md  chacha20-demo  jasmin  libjade-baseline  libjade-zeroing-loop  libjade-zeroing-unrolled  scripts  tables
$ make clean
$ find libjade-* -name "*.s" # should return nothing if the cleaning process was successful
$ find libjade-* -name "*.csv" # should return nothing as well 
$ make compile -j8 # if you have more CPU power feel free to use it
$ find libjade-* -name "*.s" | wc -l # should return 132
$ exit # if using a docker container
```

To run the benchmarks outside of the docker image, you can copy the updated directories with, for instance, the following commands:
```
$ docker cp zeroization_1:/home/zeroization/libjade-baseline . 
$ docker cp zeroization_1:/home/zeroization/libjade-zeroing-loop . 
$ docker cp zeroization_1:/home/zeroization/libjade-zeroing-unrolled .
```

Next, you can repeat the previously described steps:
```
$ make bench 
$ make compare
```

